-- The below is my attempt to translate metanet's tutorial collision code from actionscript to Lua.
-- View the original at:  http://www.metanetsoftware.com/technique/tutorialA.html
-- http://www.metanetsoftware.com/technique/tutorialB.html
-- I leave out the parts that were there to make the demo interactive.
-- Some variable names have been changed to preserve anonymity.

gamelicense = "Compatible"
require "subcritical"

local grav= 0.2 -- .3 is a bit much, .1 is a bit "on the moon"..
local drag= 0.9999999 -- 0 means full drag, 1 is no drag
local bounce= 0.3 -- must be [0,1], where 1 means full bounce. but 1 seems to incite "the flubber effect" so use 0.9 as a practical upper bound
local friction= 0.05

local min_f= 0 -- of friction
local max_f= 1

local min_b= 0 -- bounce
local max_b= 0.99

local min_g= 0 -- grav
local max_g= 1

local world_min= SubCritical.Construct("Vec2") -- these define the world bounds
local world_max= SubCritical.Construct("Vec2", 800, 800) -- my world is larger than the one in the metanet example

local objrad= 24
local tilerad= objrad*2
local objspeed= 0.2
local maxspeed= 20

-- subcritical's vector library doesn't have a scalar divide.
function vecdiv(v, s)
  return v * (1 / s)
end

-- simple physics functions

--(p.x,p.y) is projection vector, (d.x, d.y) is surface normal, tile is the tile of the world, so we can apply special friction/bounce based on the tile.
function report_collision_vs_world(this, p, d, tile)
  local v= this.pos - this.oldpos
  -- find component of velocity parallel to collision normal
  local dp= v .. d
  local n= d * dp

  local t= v - n

  -- we only want to apply collision response forces if the object is travelling into, and not out of, the collision
  local real_bounce, b, f
  if(dp < 0) then
    if(tile) then
      f= (t * tile.friction_mul * (friction + tile.friction_add))
      real_bounce= (tile.bounce_add + bounce) * tile.bounce_mul
    else
      f= t * friction
      real_bounce= 1+bounce
    end

    b= n * real_bounce
  else
    -- moving out of collision, do not apply forces
    b= sc.Construct("Vec2") -- Subcritical sets them to 0, 0 if you don't specify values
    f= sc.Construct("Vec2")
  end

  this.pos= this.pos + p -- project object out of collision
  this.oldpos= this.oldpos + p + b + f -- apply bounce+friction impulses which alter velocity
end

function integrate_verlet(this)
  local o= sc.Construct("Vec2", this.oldpos.x, this.oldpos.y) -- we can't swap buffers since mcs/sticks point directly to vector2s
  -- I have no idea what metanet means by that.
  this.oldpos= sc.Construct("Vec2", this.pos.x, this.pos.y)  -- get vector values
  -- p = position

  -- integrate
  this.pos= this.pos + ((this.oldpos * drag) - (o * drag))
  this.pos= sc.Construct("Vec2", this.pos.x, this.pos.y + grav)
end

function collide_circle_vs_world_bounds(this)
  -- collide vs x-bounds
  -- text xmin
  local d= wxmin - (this.pos.x - this.r)
  if(0 < d) then
    -- object is colliding with xmin
    report_collision_vs_world(this, {x= d, y= 0}, {x= 1, y= 0}, nil)
  else
    -- test xmax
    d= (this.pos.x + r) - wxmax
    if(0 < d) then
      --object is colliding with xmax
      report_collision_vs_world(this, {x= -d, y= 0}, {x= -1, y= 0}, nil)
    end
  end

  d= wymin - (this.pos.y - this.r)
  if(0 < d) then
    -- object is colliding with ymin
    report_collision_vs_world(this, {x= 0, y= d}, {x= 0, y= 1}, nil)
  else
    -- test ymax
    d= (this.pos.y + r) - wymax
    if(0 < d) then
      --object is colliding with ymax
      report_collision_vs_world(this, {x= 0, y= -d}, {x= 0, y= -1}, nil)
    end
  end
end

function collide_aabb_vs_world_bounds(this)
  -- collide vs. x-bounds
  -- test xmin
  local d= wxmin - (this.pos.x - this.size.x)
  if(0 < d) then
    -- object is colliding with xmin
    report_collision_vs_world(this, {x= d, y= 0}, {x= 1, y= 0}, nil)
  else
    --test xmax
    d = (this.pos.x + this.size.x) - wxmax
    if(0 < d) then
      --object is colliding with xmax
      report_collision_vs_world(this, {x= -d, y= 0}, {x= -1, y= 0}, nil)
    end
  end
  
  --collide vs. y-bounds
  --test ymin
  d= wymin - (this.pos.y - this.size.y)
  if(0 < d) then
    --object is colliding with ymin
    report_collision_vs_world(this, {x= 0, y= d}, {x= 0, y= 1}, nil)
  else
    --test ymax
    d = (this.pos.y + this.size.y) - wymax
    if(0 < d) then
      --object is colliding with ymax
      report_collision_vs_world(this, {x= 0, y= -d}, {x= 0, y= -1}, nil)
    end
  end
end


--[[
physics/collision code
this function resolves the collision between an object and a tile,
based on the tile type. (x,y) is the  projection vector.
this function returns true IF it moves the object by the specified
projection vector

NOTE: the cell offset vector (oH,oV) relative to the _tile_, i.e if the
circle is being tested against the tile to its left, oH = 1
--]]

function resolve_circle_tile(p, oh, ov, obj, tile)
  if(0 < tile.type) then
    return proj_circle_tile[tile.type](p, oh, ov, obj, tile)
  else
    trace("ResolveCircleTile() was called with an empty (or unknown) tile!: ID=" + tile.type + " ("+ tile.pos.x + "," + tile.pos.y + ")")
    return false
  end
end

function proj_circle_67_degb(x, y, oh, ov, obj, tile)
  --[[
  if we're colliding diagonally:
    -if we're in the cell pointed at by the normal, collide vs slope, else
    collide vs. the appropriate corner/vertex
  
  if obj is in this tile: collide as with aabb
  
  if obj is horiz or vertical neighbor AGAINST the slope: collide with edge
  
  if obj is vert neighb in direction of slope: collide vs. slope or vertex or halfedge
  
  if obj is horiz neighb in direction of slope: collide vs. slope or vertex
  --]]

  local tsign= tile.sign

  if(oh == 0) then
    if(ov == 0) then
      --colliding with current cell

      local ts= tile.s
      local r= obj.r

      --this gives is the coordinates of the innermost point on the AABB, relative to a point on the slope
      local o= sc.Construct("Vec2",
                                     (obj.pos.x - (ts.x * r)) - (tile.pos.x + (tsign.x * tile.size.x)),
                                     (obj.pos.y - (ts.y * r)) - (tile.pos.y - (tsign.y * tile.size.y)))
    
      --if the dotprod of (o.x,o.y) and (ts.x,ts.y) is negative, the point on the circle is in the slope
      --and we need to project it out by the magnitude of the projection of (o.x,o.y) onto (ts.x,ts.y)
      local dp= o .. ts
          
      if(dp < 0) then
        --collision; project delta onto slope and use this to displace the object
        ts= ts * -dp -- (ts.x,ts.y) is now the projection vector

        local lenN= ts:Magnitude()
        local lenP
        
        --find the smallest axial projection vector
        if(x < y) then
          --penetration in x is smaller
          lenP = x
          y = 0
          --get sign for projection along x-axis    
          if((obj.pos.x - tile.pos.x) < 0) then
            x = x * -1
          end
        else
          --penetration in y is smaller
          lenP = y
          x = 0
          --get sign for projection along y-axis    
          if((obj.pos.y - tile.pos.y)< 0) then
            y = y * -1
          end
        end

        if(lenP < lenN) then
          report_collision_vs_world(obj, sc.Construct("Vec2", x, y), sc.Construct("Vec2", x/lenP, y/lenP), tile)
          return COL_AXIS
        else
          report_collision_vs_world(obj, ts, tile.s, tile)
          return COL_OTHER
        end
      end
    else
      --colliding vertically
    
      if((tsign.y*oV) < 0) then
        --colliding with face/edge
        report_collision_vs_world(obj, sc.Construct("Vec2", 0, y*oV), sc.Construct("Vec2", 0, oV), tile)
        return COL_AXIS
      else
        --colliding with edge, slope, or vertex

        --this gives is the coordinates of the innermost point on the circle, relative to the closest tile vert
        local o= sc.Construct("Vec2",
                                       obj.pos.x - tile.pos.x,
                                       obj.pos.y - (tile.pos.y + (tsign.y * tile.size.y)))

        if((o.x * tsign.x) < 0) then
          --we're colliding with the halfface
          report_collision_vs_world(obj, sc.Construct("Vec2", 0, y*oV), sc.Construct("Vec2", 0, oV), tile)
          return COL_AXIS
        else
          --colliding with the vertex or slope

          local s= sc.Construct("Vec2", tile.s.x, tile.s.y)

          --if the component of (ox,oy) parallel to the normal's righthand normal
          --has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
          --then we project by the vertex, otherwise by the slope.
          --note that this is simply a VERY tricky/weird method of determining 
          --if the circle is in side the slope/face's voronio region, or that of the vertex.

          local perp = (o.x * -s.y) + (o.y * s.x);
          if(0 < (perp * tsign.x * tsign.y)) then
            --collide vs. vertex
            local len = o:Magnitude()
            local pen = obj.r - len
            if(0 < pen) then
              --note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
              o= vecdiv(o, len)

              report_collision_vs_world(obj, sc.Construct("Vec2", o.x * pen, o.y * pen),
                                        sc.Construct("Vec2", o.x, o.y), tile)
              return COL_OTHER
            end
          else
            --collide vs. slope
              
            --if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
            --penetrating the slope. note that this method of penetration calculation doesn't hold
            --in general (i.e it won't work if the circle is in the slope), but works in this case
            --because we know the circle is in a neighboring cell

            local dp = o .. s
            local pen = obj.r - math.abs(dp) --note: we don't need the abs because we know the dp will be positive, but just in case..
            if(0 < pen) then
              --collision; circle out along normal by penetration amount
              report_collision_vs_world(obj, s * pen, s, tile)
              return COL_OTHER
            end
          end  
        end
      end
    end
  elseif(oV == 0) then
    --colliding horizontally
      
    if((tsign.x * oH) < 0) then
      --colliding with face/edge
      report_collision_vs_world(obj, sc.Construct("Vec2", x * oH, 0), sc.Construct("Vec2", oH, 0), tile)
      return COL_AXIS
    else
      --we could only be colliding vs the slope OR a vertex
      --look at the vector form the closest vert to the circle to decide

      local slen = math.sqrt(5) --the raw slope is (-2,-1)
      local ts= sc.Construct("Vec2", tsign.x*2 / slen, tsign.y / slen) --get slope _unit_ normal; raw RH normal is (1,-2)

      --this gives is the coordinates of the innermost point on the circle, relative to the closest tile vert  
      local o= sc.Construct("Vec2", obj.pos.x - (tile.pos.x + (tsign.x * tile.size.x)),
                                     obj.pos.y - (tile.pos.y - (tsign.y * tile.size.y)))

      --if the component of (ox,oy) parallel to the normal's righthand normal
      --has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
      --then we project by the slope, otherwise by the vertex.
      --note that this is simply a VERY tricky/weird method of determining 
      --if the circle is in side the slope/face's voronio region, or that of the vertex.
      local perp= (o.x * - ts.y) + (o.y * s.x)
      if((perp * tsign.x * tsign.y) < 0) then
        --collide vs. vertex
        local len = o:Magnitude()
        local pen = obj.r - len
        if(0 < pen) then
          --note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
          o= vecdiv(o, len)

          report_collision_vs_world(obj, o * pen, o, o, tile)
          return COL_OTHER
        end
      else
        --collide vs. slope
          
        --if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
        --penetrating the slope. note that this method of penetration calculation doesn't hold
        --in general (i.e it won't work if the circle is in the slope), but works in this case
        --because we know the circle is in a neighboring cell
        local dp = o .. ts
        local pen = obj.r - math.abs(dp) --note: we don't need the abs because we know the dp will be positive, but just in case..
        if(0 < pen) then
          --collision; circle out along normal by penetration amount
          report_collision_vs_world(obj, ts * pen, tile.s, tile)
          return COL_OTHER
        end
      end
    end
  else
    --colliding diagonally
    if( 0 < ((tsign.x * oH) + (tsign.y * oV)) ) then
      --the dotprod of slope normal and cell offset is strictly positive,
      --therefore obj is in the diagonal neighb pointed at by the normal.
      
      --collide vs slope

      local ts= tile.s
  
      local r = obj.r;
      --this gives is the coordinates of the innermost point on the circle, relative to a point on the slope
      local o= sc.Construct("Vec2", (obj.pos.x - (ts.x * r)) - (tile.pos.x + (tsign.x * tile.size.x)),
                                     (obj.pos.y - (ts.y * r)) - (tile.pos.y - (tsign.y * tile.size.y)))
    
      --if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
      --and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      local dp = o .. ts
          
      if(dp < 0) then
        --collision; project delta onto slope and use this to displace the object  
        --(sx,sy)*-dp is the projection vector

        report_collision_vs_world(obj, -ts * dp, tile.s, tile)
        return COL_OTHER
      end
      return COL_NONE
    else
      --collide vs the appropriate vertex
      local v= tile.pos + sc.Construct("Vec2", oH * tile.size.x, oV * tile.size.y)

      local d= obj.pos - v --calc vert->circle vector

      local len= d:Magnitude()
      local pen = obj.r - len;
      if(0 < pen) then
        --vertex is in the circle; project outward
        if(len == 0) then
          --project out by 45deg
          d= sc.Construct("Vec2", oH / math.sqrt(2), oV / math.sqrt(2))
        else
          d= vecdiv(d, len)
        end

        report_collision_vs_world(obj, d * pen, d, tile)
        return COL_OTHER
      end
    end
  end
  return COL_NONE
end

function proj_circle_67_degs(x, y, oH, oV, obj, tile)
  --if the object is in a cell pointed at by signx, no collision will ever occur
  --otherwise,
  --
  --if we're colliding diagonally:
  --  -collide vs. the appropriate vertex
  --if obj is in this tile: collide vs slope or vertex or axis
  --if obj is vert neighb in direction of slope: collide vs. slope or vertex
  --if obj is vert neighb against the slope:
  --   if(distance in y from circle to 90deg corner of tile < 1/2 tileheight, collide vs. face)
  --   else(collide vs. corner of slope) (vert collision with a non-grid-aligned vert)
  --if obj is horiz neighb against direction of slope: collide vs. face

  local tsign= tile.sign

  if(0 < (tsign.x * oH)) then
    --object will never collide vs tile, it can't reach that far
    return COL_NONE
  elseif(oH == 0) then
    if(oV == 0) then
      --colliding with current tile
      --we could only be colliding vs the slope OR a vertex
      --look at the vector form the closest vert to the circle to decide

      local ts= tile.s
      
      local r = obj.r
      --this gives is the coordinates of the innermost point on the circle, relative to the tile corner  
      local o= sc.Construct("Vec2", obj.pos.x - tile.pos.x, obj.pos.y - (tile.pos.y - (tsign.y * tile.size.y)))
    
      --if the component of (ox,oy) parallel to the normal's righthand normal
      --has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
      --then we project by the normal or axis, otherwise by the corner/vertex
      --note that this is simply a VERY tricky/weird method of determining 
      --if the circle is in side the slope/face's voronoi region, or that of the vertex.
        
      local perp = (o.x* -ts.y) + (o.y * ts.x)
      if((perp * tsign.x * tsign.y) < 0) then
        --collide vs. vertex
        local len = o:Magnitude()
        local pen = r - len
        if(0 < pen) then
          --note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
          o= vecdiv(o, len)
          report_collision_vs_world(obj, o * pen, o, tile)
          return COL_OTHER
        end
      else
        --collide vs. slope or vs axis
        o= o - (ts * r) --this gives us the vector from a point on the slope to the innermost point on the circle
    
        --if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
        --and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
        local dp = o .. ts
        
        if(dp < 0) then
          --collision project delta onto slope and use this to displace the object
          ts= ts * -dp --(sx,sy) is now the projection vector
            
          local lenN = ts:Magnitude()
      
          --find the smallest axial projection vector
          if(x < y) then
            --penetration in x is smaller
            lenP = x
            y = 0  
            --get sign for projection along x-axis    
            if((obj.pos.x - tile.pos.x) < 0) then
              x= x * -1
            end
          else
            --penetration in y is smaller
            lenP = y
            x = 0  
            --get sign for projection along y-axis    
            if((obj.pos.y - tile.pos.y)< 0) then
              y = y * -1
            end
          end

          if(lenP < lenN) then
            report_collision_vs_world(obj, sc.Construct("Vec2", x, y), sc.Construct("Vec2", x/lenP, y/lenP), tile)
            return COL_AXIS
          else
            report_collision_vs_world(obj, ts, tile.s, tile)
            return COL_OTHER
          end
        end
      end
    else
      --colliding vertically
      
      if((tsign.y * oV) < 0) then
        --colliding with face/edge OR with corner of wedge, depending on our position vertically
          
        --collide vs. vertex
        --get diag vertex position
        local v= sc.Construct("Vec2", tile.pos.x, tile.pos.y - (tsign.y * tile.size.y))

        local d= obj.pos - v --calc vert->circle vector

        if((d.x * tsign.x) < 0) then
          --colliding vs face
          report_collision_vs_world(obj, sc.Construct("Vec2", 0, y * oV), sc.Construct("Vec2", 0, oV), tile)
          return COL_AXIS
        else
          --colliding vs. vertex
            
          local len = d:Magnitude()
          local pen = obj.r - len
          if(0 < pen) then
            --vertex is in the circle project outward
            if(len == 0) then
              --project out by 45deg
              d= sc.Construct("Vec2", oH / math.sqrt(2), oV / math.sqrt(2))
            else
              d= vecdiv(d, len)
            end
            report_collision_vs_world(obj, d * pen, d, tile)
            return COL_OTHER
          end
        end
      else
        --we could only be colliding vs the slope OR a vertex
        --look at the vector form the closest vert to the circle to decide

        local ts= tile.s

        --this gives is the coordinates of the innermost point on the circle, relative to the closest tile vert  
        local o= sc.Construct("Vec2", obj.pos.x - (tile.pos.x - (tsign.x * tile.size.x)),
                                       obj.pos.y - (tile.pos.y + (oV * tile.size.y)))
    
        --if the component of (ox,oy) parallel to the normal's righthand normal
        --has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
        --then we project by the vertex, otherwise by the normal.
        --note that this is simply a VERY tricky/weird method of determining 
        --if the circle is in side the slope/face's voronio region, or that of the vertex.                        
        local perp = (o.x * -ts.y) + (o.y * ts.x)
        if(0 < (perp * tsign.x * tsign.y)) then
          --collide vs. vertex
          local len = o:Magnitude()
          local pen = obj.r - len
          if(0 < pen) then
            --note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
            o= vecdiv(o, len)
            report_collision_vs_world(obj, o * pen, o, tile)
            return COL_OTHER
          end
        else
          --collide vs. slope
              
          --if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
          --penetrating the slope. note that this method of penetration calculation doesn't hold
          --in general (i.e it won't work if the circle is in the slope), but works in this case
          --because we know the circle is in a neighboring cell
          local dp = o .. ts
          local pen = obj.r - math.abs(dp)--note: we don't need the abs because we know the dp will be positive, but just in case..        

          if(0 < pen) then
            --collision circle out along normal by penetration amount
            report_collision_vs_world(obj, ts * pen, tile.s, tile)
            return COL_OTHER
          end
        end
      end
    end
  elseif(oV == 0) then
    --colliding horizontally we can assume that (signy*oV) < 0
    --due to the first conditional far above

    report_collision_vs_world(obj, sc.Construct("Vec2", x*oH, 0), sc.Construct("Vec2", oH, 0), tile)
    
    return COL_AXIS
  else
    --colliding diagonally due to the first conditional above,
    --obj is vertically offset against slope, and offset in either direction horizontally

    --collide vs. vertex
    --get diag vertex position
    local v= sc.Construct("Vec2", tile.pos.x + (oH * tile.size.x), tile.pos.y + (oV * tile.size.y))
    local d= obj.pos - v --calc vert->circle vector
      
    local len = d:Magnitude()
    local pen = obj.r - len
    if(0 < pen) then
      --vertex is in the circle project outward
      if(len == 0) then 
        --project out by 45deg
        d= sc.Construct("Vec2", oH / math.sqrt(2), oV / math.sqrt(2))
      else
        d= vecdiv(d, len)
      end
      report_collision_vs_world(obj, d * pen, d, tile)
      return COL_OTHER
    end
  end
  return COL_NONE
end

function proj_circle_22_degb(x, y, oH, oV, obj, tile)
  --if we're colliding diagonally:
  --  -if we're in the cell pointed at by the normal, collide vs slope, else
  --  collide vs. the appropriate corner/vertex
  --
  --if obj is in this tile: collide as with aabb
  --
  --if obj is horiz or vertical neighbor AGAINST the slope: collide with edge
  --
  --if obj is horiz neighb in direction of slope: collide vs. slope or vertex or edge
  --
  --if obj is vert neighb in direction of slope: collide vs. slope or vertex

  local tsign= tile.sign

  if(oH == 0) then
    if(oV == 0) then
      --colliding with current cell

      local ts= tile.s

      local r = obj.r
      --this gives is the coordinates of the innermost point on the AABB, relative to a point on the slope
      local o= sc.Construct("Vec2", (obj.pos.x - (ts.x * r)) - (tile.pos.x - (tsign.x * tile.size.x)),
                                     (obj.pos.y - (ts.y * r)) - (tile.pos.y + (tsign.y * tile.size.y)))
    
      --if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
      --and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      local dp = o .. s
          
      if(dp < 0) then
        --collision project delta onto slope and use this to displace the object
        ts= ts * -dp --(sx,sy) is now the projection vector
              
        local lenN = s:Magnitude()
        
        --find the smallest axial projection vector
        if(x < y) then
          --penetration in x is smaller
          lenP = x
          y = 0  
          --get sign for projection along x-axis    
          if((obj.pos.x - tile.pos.x) < 0) then
            x = x * -1
          end
        else
          --penetration in y is smaller
          lenP = y
          x = 0  
          --get sign for projection along y-axis    
          if((obj.pos.y - tile.pos.y)< 0) then
            y = y * -1
          end
        end
  
        if(lenP < lenN) then
          report_collision_vs_world(obj, sc.Construct("Vec2", x, y), sc.Construct("Vec2", x/lenP, y/lenP), tile)
          return COL_AXIS
        else
          report_collision_vs_world(obj, ts, tile.s, tile)
          return COL_OTHER
        end
      end
    else
      --colliding vertically
      
      if((tsign.y * oV) < 0) then
        --colliding with face/edge
        report_collision_vs_world(obj, sc.Construct("Vec2", 0, y * oV), sc.Construct("Vec2", 0, oV), tile)
        return COL_AXIS
      else
        --we could only be colliding vs the slope OR a vertex
        --look at the vector form the closest vert to the circle to decide

        local ts= tile.s

        --this gives is the coordinates of the innermost point on the circle, relative to the closest tile vert  
        local o= sc.Construct("Vec2", obj.pos.x - (tile.pos.x - (tsign.x * tile.size.x)),
                              obj.pos.y - (tile.pos.y - (tsign.y * tile.size.y)))

        --if the component of (ox,oy) parallel to the normal's righthand normal
        --has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
        --then we project by the vertex, otherwise by the normal.
        --note that this is simply a VERY tricky/weird method of determining 
        --if the circle is in side the slope/face's voronio region, or that of the vertex.
        local perp = (o.x * -ts.y) + (o.y * ts.x)
        if(0 < (perp * tsign.x * tsign.y)) then
          --collide vs. vertex
          local len = o:Magnitude()
          local pen = obj.r - len
          if(0 < pen) then
            --note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
            o= vecdiv(o, len)
            report_collision_vs_world(obj, o * pen, o, tile)
            return COL_OTHER
          end
        else
          --collide vs. slope
          
          --if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
          --penetrating the slope. note that this method of penetration calculation doesn't hold
          --in general (i.e it won't work if the circle is in the slope), but works in this case
          --because we know the circle is in a neighboring cell
          local dp = o .. s
          local pen = obj.r - math.abs(dp)--note: we don't need the abs because we know the dp will be positive, but just in case..
          if(0 < pen) then
            --collision circle out along normal by penetration amount
            report_collision_vs_world(obj, ts * pen, ts, tile)
            return COL_OTHER
          end
        end
      end
    end
  elseif(oV == 0) then
    --colliding horizontally
    
    if((tsign.x * oH) < 0) then
      --colliding with face/edge
      report_collision_vs_world(obj, sc.Construct("Vec2", x * oH, 0), sc.Construct("Vec2", oH, 0), tile)
      return COL_AXIS
    else
      --colliding with edge, slope, or vertex

      --this gives is the coordinates of the innermost point on the circle, relative to the closest tile vert
      local o= sc.Construct("Vec2", obj.pos.x - (tile.pos.x + (tsign.x * tile.size.x)),
                            obj.pos.y - tile.pos.y)
        
      if((o.y * tsign.y) < 0) then
        --we're colliding with the halfface
        report_collision_vs_world(obj, sc.Construct("Vec2", x * oH, 0), sc.Construct("Vec2", oH, 0), tile)
        return COL_AXIS      
      else
        --colliding with the vertex or slope

        local ts= tile.s
                
        --if the component of (ox,oy) parallel to the normal's righthand normal
        --has the same sign as the slope of the slope (the sign of the slope's slope is signx*signy)
        --then we project by the slope, otherwise by the vertex.
        --note that this is simply a VERY tricky/weird method of determining 
        --if the circle is in side the slope/face's voronio region, or that of the vertex.
        local perp = (o.x * -ts.y) + (o.y * ts.x)
        if((perp * tsign.x * tsign.y) < 0) then
          --collide vs. vertex
          local len = o:Magnitude()
          local pen = obj.r - len
          if(0 < pen) then
            --note: if len=0, then perp=0 and we'll never reach here, so don't worry about div-by-0
            o= vecdiv(o, len)
            report_collision_vs_world(obj, o * pen, o, tile)
            return COL_OTHER
          end
        else
          --collide vs. slope
            
          --if the component of (ox,oy) parallel to the normal is less than the circle radius, we're
          --penetrating the slope. note that this method of penetration calculation doesn't hold
          --in general (i.e it won't work if the circle is in the slope), but works in this case
          --because we know the circle is in a neighboring cell
          local dp = o .. ts
          local pen = obj.r - math.abs(dp)--note: we don't need the abs because we know the dp will be positive, but just in case..
          if(0 < pen) then
            --collision circle out along normal by penetration amount
            report_collision_vs_world(obj, ts * pen, tile.s, tile)
            return COL_OTHER
          end
        end
      end
    end
  else
    --colliding diagonally
    if( 0 < ((tsign.x * oH) + (tsign.y * oV)) ) then
      --the dotprod of slope normal and cell offset is strictly positive,
      --therefore obj is in the diagonal neighb pointed at by the normal.
      
      --collide vs slope

      --we should really precalc this at compile time, but for now, fuck it
      local slen = math.sqrt(5)--the raw slope is (-2,-1)
      local ts= sc.Construct("Vec2", tsign.x / slen, (tsign.y * 2) / slen) --get slope _unit_ normal raw RH normal is (1,-2)
  
      local r = obj.r
      --this gives is the coordinates of the innermost point on the circle, relative to a point on the slope
      local o= sc.Construct("Vec2", (obj.pos.x - (ts.x * r)) - (tile.pos.x - (tsign.x * tile.size.x)),
                            (obj.pos.y - (ts.y * r)) - (tile.pos.y + (tsign.y * tile.size.y)))
    
      --if the dotprod of (ox,oy) and (sx,sy) is negative, the point on the circle is in the slope
      --and we need toproject it out by the magnitude of the projection of (ox,oy) onto (sx,sy)
      local dp = o .. s
      if(dp < 0) then
        --collision project delta onto slope and use this to displace the object  
        --(sx,sy)*-dp is the projection vector
        report_collision_vs_world(obj, ts * -1 * dp, tile.s, tile)
        return COL_OTHER
      end
      return COL_NONE
    else
      --collide vs the appropriate vertex
      local v= sc.Construct("Vec2", tile.pos.x + (oH * tile.size.x), tile.pos.y + (oV * tile.size.y))

      local d= obj.pos - v --calc vert->circle vector
      
      local len = d:Magnitude()
      local pen = obj.r - len
      if(0 < pen) then
        --vertex is in the circle project outward
        if(len == 0) then
          --project out by 45deg
          d= sc.Construct("Vec2", oH / math.sqrt(2), oV / math.sqrt(2))
        else
          d= vecdiv(d, len)
        end
        
        report_collision_vs_world(obj, d * pen, d, tile)
        return COL_OTHER
      end
    end
  end
  return COL_NONE
end
