--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "io"

require "quantum"

require "collision"
require "shot"

local gt_key_down_funcs= {}
local gt_key_up_funcs= {}

local travel_width= 512
local travel_height= 512
local travel_offset= sc.Construct("Vec2", 0, screen_height - travel_height)
local travel_xoff= travel_offset.x
local travel_yoff= travel_offset.y

local minimap_width= 256
local minimap_height= 256
local minimap_xoff= travel_width + 16
local minimap_yoff= 16

active_rooms= {}

local profile_section_data= {}

local tick_time_total= 0
local ticks_passed= 0
local os_clock_detected_load_spikes= 0
local gettime_detected_load_spikes= 0
local matched_spikeds= 0

local unrot_graphic= sc.Construct("Graphic", travel_width, travel_height)
local minimap_graphic= sc.Construct("Graphic", minimap_width, minimap_height)

function profile_section_start(name)
  if not profile_section_data[name] then
    profile_section_data[name]= {time_total= 0, ticks= 0, time_max= 0}
  end
  profile_section_data[name].tick_start= SCUtil.GetTime()
end

local elapsed_time= 0

function profile_section_end(name)
  local tick_end= SCUtil.GetTime()
  elapsed_time= tick_end - profile_section_data[name].tick_start
  profile_section_data[name].time_total= profile_section_data[name].time_total + elapsed_time
  if elapsed_time > profile_section_data[name].time_max then
    profile_section_data[name].time_max= elapsed_time
  end
  profile_section_data[name].ticks= profile_section_data[name].ticks+1
end

function profile_data_print()
  for key, value in pairs(profile_section_data) do
    print(key.." timing data:")
    print("Average:  "..value.time_total / value.ticks)
    print("Max:      "..value.time_max)
  end
end

function gt_start()
--  print("Back so soon?")
  gt_ticks= SCUtil.GetTime() * 30
  gt_rticks= 0
  gt_should_exit= nil
  screen:SetPrimitiveColor(0, 0, 0)
  screen:DrawRect(0, 0, screen_width, screen_height)
  gt_key_down_funcs["left"]= gt_left_down
  gt_key_down_funcs["right"]= gt_right_down
  gt_key_down_funcs["up"]= gt_up_down
  gt_key_down_funcs["down"]= gt_down_down
  gt_key_down_funcs["escape"]= gt_escape_down
  gt_key_down_funcs["p"]= gt_p_down
  gt_key_down_funcs["z"]= gt_z_down
  gt_key_down_funcs["m"]= gt_m_down
  gt_key_down_funcs["f"]= gt_f_down
  gt_key_down_funcs["r"]= gt_r_down
  gt_key_down_funcs["c"]= gt_c_down
  gt_key_down_funcs["left shift"]= gt_return_down
  gt_key_down_funcs["space"]= gt_space_down
  gt_key_down_funcs["/"]= gt_screenshot_down
  gt_key_up_funcs["d"]= gt_d_up
  gt_key_up_funcs["z"]= gt_z_up
  gt_key_up_funcs["left"]= gt_left_up
  gt_key_up_funcs["right"]= gt_right_up
  gt_key_up_funcs["up"]= gt_up_up
  gt_key_up_funcs["down"]= gt_down_up
  gt_key_up_funcs["left shift"]= gt_return_up
  check_room_active(player.cur_room, player)
  print("Calling quantum.run")
  collectgarbage("collect")
  -- The garbage collector causes a pause in the game every time it runs a cycle.
  -- gt_idle steps it a bit each tick so we don't run out of memory.
  collectgarbage("stop")
  quantum.run({tickrate= 30, process= gt_idle, handle_event= gt_event_handle, render= gt_display})
  print("Returned from quantum.run")
end

function gt_cleanup()
  print("gt_cleanup")
  profile_data_print()
  print("os_clock_detected_load_spikes:  "..os_clock_detected_load_spikes)
  print("gettime_detected_load_spikes:  "..gettime_detected_load_spikes)
  print("matched_spikeds:  "..matched_spikeds)
  gt_should_exit= true
end

function check_room_active(rid, avatar)
  if(not active_rooms[rid]) then
    load_room(rid)
  end
  if avatar.name == player.name then -- Seems like the easiest way to check if the avatar is the player
    minimap_graphic= SCUtil.ScaleFast(load_img_managed("rooms/"..rid..".png"), minimap_width, minimap_height)
  end
  assert(active_rooms[rid])
  assert(active_rooms[rid].avatars)
  active_rooms[rid].avatars[avatar.name]= avatar
end

function avatar_moved_from_room(avatar, from, dir)
  local froom= active_rooms[from]
  froom.avatars[avatar.name]= nil
  check_room_active(froom.conns[dir].side, avatar)
  --print("Avatar " .. avatar.name .. " moved from room " .. from .. " through " .. dir .. " side.")
  actor.add_rot(avatar, dir, froom.conns[dir].rot)
  if(not next(froom.avatars)) then
    unload_room(from)
  end
  avatar.cur_room= froom.conns[dir].side
  if(avatar.ai_room_change_func) then
    avatar.ai_room_change_func(avatar)
  end
end

function kill_avatar(ava)
  local avcr= ava.cur_room
  local avroom= active_rooms[avcr]
  avroom.avatars[ava.name]= nil
  if not next(avroom.avatars) then
    active_rooms[avcr]= nil
  end
  print(ava.name .. " has been killed.")
  if ava == player then
    print("You died.")
    next_play_mode= "quit"
    gt_cleanup()
  end
end

function calc_transform_data()
  local pcr= player.cur_rot
  local ppos= player.pos
  local ox= ppos.x - (travel_width / 2)
  local oy= ppos.y - (travel_height / 2)
  if ox < 0 then
    ox= 0
  end
  if ox + travel_width > world_max.x then
    ox= world_max.x - travel_width
  end
  if oy < 0 then
    oy= 0
  end
  if oy + travel_height > world_max.y then
    oy= world_max.y - travel_height
  end
  return sc.Construct("Vec2", ox, oy)
end

function point_transform(o, p)
  return (p - o)
end

function coord_transform(o, x, y)
  return (x - o.x), (y - o.y)
end

local blink_cycle_mod= 0
local blink_cycle_inc= .125

local edge_coords= {}
local edge_indices= {}

local gun_coords_a, gun_indices_a= {}, {}
local gun_coords_b= {}
local gun_coords_c= {}
local gun_coords_d= {}
local firing_vec_coords= {}
local firing_vec_indices= {}

function gt_display()
  local droom= active_rooms[player.cur_room]
  local droug= unrot_graphic
  local drommg= minimap_graphic
  local origin= calc_transform_data()
  local orix= origin.x
  local oriy= origin.y
  local vis_xmin, vis_ymin= orix, oriy
  local vis_xmax, vis_ymax= orix + travel_width, oriy + travel_height
  local droom_sizex= droom.size.x
  edge_coords= {}
  edge_indices= {}
  function draw_wall(x, y, other)
    local wid= pos_to_wid(droom_sizex, x, y)
    local wall= droom.walls[wid]
    if wall then
    --[[
      for side, edge in pairs(wall.edges) do
        add_point_to_coords(edge_coords, edge_indices, point_transform(origin, edge.a))
        add_point_to_coords(edge_coords, edge_indices, point_transform(origin, edge.b))
      end
      ]]
      local wallpos= wall.pos
      local wallsize= wall.size
      local x, y= coord_transform(origin, wallpos.x - wallsize.x, wallpos.y - wallsize.y)
      droug:Blit(brick_grc, x, y)
    end
  end
  spaced_area_iterate(vis_xmin, vis_xmax, vis_ymin, vis_ymax, 32, draw_wall)
  for name, place in pairs(droom.places) do
    local x, y= coord_transform(origin, place.pos.x - place.size.x, place.pos.y - place.size.y)
    if not (x < -64 or y < -64 or x > travel_width or y > travel_height) then
      droug:Blit(place_grc, place.rot*64, place.gin*64, 64, 64, x, y)
    end
  end
  for id, shot in pairs(droom.shots) do
    shot.draw_func(droug, origin, shot)
  end
  gun_coords_a, gun_indices_a= {}, {}
  gun_coords_b= {}
  gun_coords_c= {}
  gun_coords_d= {}
  firing_vec_coords= {}
  firing_vec_indices= {}
  for name, ava in pairs(droom.avatars) do
    local avag_pos= ava.g_pos
    local avasize= ava.size
    local blit_x, blit_y= coord_transform(origin, avag_pos.x - avasize.x, avag_pos.y - avasize.y)
    local avacolor= ava.color
    local abg_frisket= ava.bg_frisket
    droug:SetPrimitiveColor(0, 0, 0)
    droug:BlitFrisket(abg_frisket, blit_x, blit_y)
    droug:SetPrimitiveColor(avacolor[1], avacolor[2], avacolor[3])
    droug:BlitFrisket(ava.color_frisket, blit_x, blit_y)
    if ava.was_hit then
      if ava.ship.shield > 0 then
        droug:SetPrimitiveColor(0, 1, .5, .25)
      else
        droug:SetPrimitiveColor(1, .5, 0, .25)
      end
      droug:BlitFrisket(abg_frisket, blit_x, blit_y)
      ava.was_hit= ava.was_hit - 1
      if ava.was_hit <= 0 then
        ava.was_hit= nil
      end
    end
    for id, gun in pairs(ava.ship.primary_guns) do
      -- Setting avasta cuts out two hash table lookups.
      -- Setting avadir cuts out one hash table lookup.
      local avasta= ava.anim_info.status
      local avadir= avasta.dir
      local avaori= avasta.orientation
      local real_aim= dir_aim_flip[avadir](gun.aim) + ori_to_aim_offset_table[avaori]
      local aim_vec= sc.Construct("Vec2", math.cos(real_aim), math.sin(real_aim))
      local barrel_vec= aim_vec * 31
      local perp_aim= real_aim + (math.pi * .5)
      local aim_perp_vec= sc.Construct("Vec2", math.cos(perp_aim), math.sin(perp_aim))
      local real_pos= ava.pos + rot_point_cen_funcs[ori_to_ind[avaori]](dir_offset_flip[avadir](gun.offset))
      local sa= real_pos + (aim_perp_vec * 3.5)
      local sb= real_pos + (aim_perp_vec * 2.5)
      local sc= real_pos + (aim_perp_vec * 1.5)
      local sd= real_pos + (aim_perp_vec * .5)
      local se= real_pos + (aim_perp_vec * -.5)
      local sf= real_pos + (aim_perp_vec * -1.5)
      local sg= real_pos + (aim_perp_vec * -2.5)
      local sh= real_pos + (aim_perp_vec * -3.5)
      add_point_to_coords(gun_coords_a, gun_indices_a, point_transform(origin, sa))
      add_point_to_coords(gun_coords_a, gun_indices_a, point_transform(origin, sa+barrel_vec))
      add_point_to_coords(gun_coords_a, gun_indices_a, point_transform(origin, sh))
      add_point_to_coords(gun_coords_a, gun_indices_a, point_transform(origin, sh+barrel_vec))
      add_point_only_to_coords(gun_coords_b, point_transform(origin, sb))
      add_point_only_to_coords(gun_coords_b, point_transform(origin, sb+barrel_vec))
      add_point_only_to_coords(gun_coords_b, point_transform(origin, sg))
      add_point_only_to_coords(gun_coords_b, point_transform(origin, sg+barrel_vec))
      add_point_only_to_coords(gun_coords_c, point_transform(origin, sc))
      add_point_only_to_coords(gun_coords_c, point_transform(origin, sc+barrel_vec))
      add_point_only_to_coords(gun_coords_c, point_transform(origin, sf))
      add_point_only_to_coords(gun_coords_c, point_transform(origin, sf+barrel_vec))
      add_point_only_to_coords(gun_coords_d, point_transform(origin, sd))
      add_point_only_to_coords(gun_coords_d, point_transform(origin, sd+barrel_vec))
      add_point_only_to_coords(gun_coords_d, point_transform(origin, se))
      add_point_only_to_coords(gun_coords_d, point_transform(origin, se+barrel_vec))

      --local firing_vec= calc_firing_vec(avadir, avaori, gun.aim, 0, gun.muzzle_velocity, ava.speed)
      --add_point_to_coords(firing_vec_coords, firing_vec_indices, point_transform(origin, real_pos))
      --add_point_to_coords(firing_vec_coords, firing_vec_indices, point_transform(origin, real_pos + firing_vec))
    end
  end
  droug:SetPrimitiveColor(0, 0, 0)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_a), SCUtil.CompileIndices(gun_indices_a), 1.1)
  droug:SetPrimitiveColor(.125, .125, .125)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_b), SCUtil.CompileIndices(gun_indices_a), 1.1)
  droug:SetPrimitiveColor(.25, .25, .25)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_c), SCUtil.CompileIndices(gun_indices_a), 1.1)
  droug:SetPrimitiveColor(.5, .5, .5)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_d), SCUtil.CompileIndices(gun_indices_a), 1.1)
  --[[
  droug:SetPrimitiveColor(1, 1, 0)
  local px, py= coord_transform(origin, player.pos.x, player.pos.y)
  local psx, psy= coord_transform(origin, player.pos.x + player.speed.x, player. pos.y + player.speed.y)
  droug:DrawLines(SCUtil.CompileCoords({px, py, psx, psy}),
                                        SCUtil.CompileIndices({0, 1}), 2)
  droug:SetPrimitiveColor(0, 1, 1)
  local psax, psay= coord_transform(origin, player.pos.x + player.speed.x + player.speed_adjust.x,
                                         player. pos.y + player.speed.y + player.speed_adjust.y)
  droug:DrawLines(SCUtil.CompileCoords({px, py, psax, psay}), SCUtil.CompileIndices({0, 1}), 2)
  if(collision_acoords[1]) then
    for n= 1, #collision_acoords, 1 do
      collision_acoords[n]= point_transform(origin, collision_acoords[n])
    end
    droug:SetPrimitiveColor(1, 0, 0)
    droug:DrawLines(SCUtil.CompileCoords(collision_acoords), SCUtil.CompileIndices(collision_aindices), 2)
  end
  if(collision_bcoords[1]) then
    for n= 1, #collision_bcoords, 1 do
      collision_bcoords[n]= point_transform(origin, collision_bcoords[n])
    end
    droug:SetPrimitiveColor(0, 1, 0)
    droug:DrawLines(SCUtil.CompileCoords(collision_bcoords), SCUtil.CompileIndices(collision_bindices), 2)
  end
  if(collision_ccoords[1] and collision_cindices[1]) then
    for n= 1, #collision_ccoords, 1 do
      collision_ccoords[n]= point_transform(origin, collision_ccoords[n])
    end
    droug:SetPrimitiveColor(0, 0, 1)
    droug:DrawLines(SCUtil.CompileCoords(collision_ccoords), SCUtil.CompileIndices(collision_cindices), 2)
  end
  if firing_vec_coords[1] then
    for n= 1, #firing_vec_coords, 1 do
      firing_vec_coords[n]= point_transform(origin, firing_vec_coords[n])
    end
    droug:SetPrimitiveColor(1, 0, 0, .5)
    droug:DrawLines(SCUtil.CompileCoords(firing_vec_coords), SCUtil.CompileIndices(firing_vec_indices), 2)
    firing_vec_coords= {}
    firing_vec_indices= {}
  end
  if firea_coords[1] then
    for n= 1, #firea_coords, 1 do
      firea_coords[n]= point_transform(origin, firea_coords[n])
    end
    droug:SetPrimitiveColor(0, 1, 0, .5)
    droug:DrawLines(SCUtil.CompileCoords(firea_coords), SCUtil.CompileIndices(firea_indices), 2)
    firea_coords= {}
    firea_indices= {}
  end
  if fireb_coords[1] then
    for n= 1, #fireb_coords, 1 do
      fireb_coords[n]= point_transform(origin, fireb_coords[n])
    end
    droug:SetPrimitiveColor(0, 0, 1, .5)
    droug:DrawLines(SCUtil.CompileCoords(fireb_coords), SCUtil.CompileIndices(fireb_indices), 2)
    fireb_coords= {}
    fireb_indices= {}
  end
  if(edge_coords[1] and edge_indices[1]) then
    droug:SetPrimitiveColor(1, 0, 1)
    droug:DrawLines(SCUtil.CompileCoords(edge_coords), SCUtil.CompileIndices(edge_indices), 2)
  end
  droug:SetPrimitiveColor(1, 1, 1)
  local brx, bry= coord_transform(origin, player.below_rect.pos.x - player.below_rect.size.x, player.below_rect.pos.y - player.below_rect.size.y)
  droug:DrawBox(brx, bry, player.below_rect.size.x * 2, player.below_rect.size.y * 2)
  ]]
  local pcr= player.cur_rot
  if(pcr == 0) then
    screen:Blit(droug, travel_xoff, travel_yoff)
    screen:Blit(drommg, minimap_xoff, minimap_yoff)
  elseif pcr == 1 then
    screen:Blit(SCUtil.RotateRight(droug), travel_xoff, travel_yoff)
    screen:Blit(SCUtil.RotateRight(drommg), minimap_xoff, minimap_yoff)
  elseif pcr == 2 then
    screen:Blit(SCUtil.Flip(droug), travel_xoff, travel_yoff)
    screen:Blit(SCUtil.Flip(drommg), minimap_xoff, minimap_yoff)
  elseif pcr == 3 then
    screen:Blit(SCUtil.RotateLeft(droug), travel_xoff, travel_yoff)
    screen:Blit(SCUtil.RotateLeft(drommg), minimap_xoff, minimap_yoff)
  end
  for name, ava in pairs(droom.avatars) do
    local avapos= rot_point_funcs[pcr](ava.pos, world_max)
    local avasize= ava.size
    local avasizex= avasize.x
    local avasizey= avasize.y
    local avacolor= ava.color
    if pcr == 1 or pcr == 3 then
      avasizex, avasizey= avasizey, avasizex
    end
    local mini_blit_x= math.floor((avapos.x - avasizex) / 16) + minimap_xoff
    local mini_blit_y= math.floor((avapos.y - avasizey) / 16) + minimap_yoff
    screen:SetPrimitiveColor(avacolor[1] * blink_cycle_mod, avacolor[2] * blink_cycle_mod, avacolor[3] * blink_cycle_mod)
    screen:DrawRect(mini_blit_x, mini_blit_y, avasizex / 8, avasizey / 8)
  end
  screen:Update()
  if should_screenshot then
    local dump_img= screen:TakeSnapshot()
    local handle, message= io.open(SCPath("tint_screenie.png"), "wb")
    if not handle then
      print("Could not dump screenshot, failed to open file:  " .. mes)
    else
      pngdump:Dump(dump_img, handle, handle.write)
      handle:close()
      print("Dumped screenshot.")
    end
    should_screenshot= nil
  end
  droug:SetPrimitiveColor(0, 0, 0)
  droug:DrawRect(0, 0, travel_width, travel_height)
  screen:SetPrimitiveColor(0, 0, 0)
  -- The blinking dots of the avatars can stick outside the minimap a bit, so they need to be undrawn.
  screen:DrawBox(minimap_xoff-2, minimap_yoff-2, minimap_width+4, minimap_height+4, 2)
  blink_cycle_mod= blink_cycle_mod + blink_cycle_inc
  if blink_cycle_mod > 1 then
    blink_cycle_inc= -.125
  end
  if blink_cycle_mod < .25 then
    blink_cycle_mod= .25
    blink_cycle_inc= .125
  end
  -- Old undraw code for when the screen wasn't centered on the player.
  --[[
  for id, shot in pairs(droom.shots) do
    shot.undraw_func(droug, origin, shot)
  end
  for name, ava in pairs(droom.avatars) do
    local gpx, gpy= coord_transform(origin, ava.g_pos.x - ava.size.x, ava.g_pos.y - ava.size.y)
    droug:DrawRect(gpx, gpy, 64, 64)
  end
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_a), SCUtil.CompileIndices(gun_indices_a), 1.1)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_b), SCUtil.CompileIndices(gun_indices_a), 1.1)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_c), SCUtil.CompileIndices(gun_indices_a), 1.1)
  droug:DrawLines(SCUtil.CompileCoords(gun_coords_d), SCUtil.CompileIndices(gun_indices_a), 1.1)
  -- brx and bry were set when the rect was drawn above
  droug:DrawBox(brx, bry, player.below_rect.size.x * 2, player.below_rect.size.y * 2)
  ]]
  return nil
end

function gt_idle()
  tick_start_time= os.clock()
  profile_section_start("gt_idle")
  if(not collision_pause) then
    local avs_that_moved_from_room= {}
    for rid, croom in pairs(active_rooms) do
      --profile_section_start("avcycle")
      for avn, ava in pairs(croom.avatars) do
        if(ava.ai_think_func) then
          --profile_section_start("ai_think")
          ava.ai_think_func(ava, gt_rticks, croom)
          --profile_section_end("aithink")
        end
        --profile_section_start("av_tickle")
        actor.tickly_actions(ava)
        --profile_section_end("av_tickle")
        --profile_section_start("av_coll")
        coll_check_obj(ava, croom)
        --profile_section_end("av_coll")
        --profile_section_start("av_update_pos")
        actor.update_pos(ava)
        --profile_section_end("av_update_pos")
        --profile_section_start("av_check_in_room")
        avs_that_moved_from_room[avn]= actor.check_in_room(ava)
        --profile_section_end("av_check_in_room")
      end
      --profile_section_end("avcycle")
      -- Yeah, I know it's imperfect collision checking to do the shots and the players separately, but it's much simpler this way.
      --profile_section_start("shot_cycle")
      for id, shot in pairs(croom.shots) do
        shot.guide_func(shot)
        coll_check_obj(shot, croom)
        if(not update_shot(shot)) then -- update_shot returns nil if the shot needs to be destroyed
          croom.shots[id]= nil
        end
        for avn, ava in pairs(croom.avatars) do
          if avn ~= shot.team then
            if rects_overlap_aabb(shot, ava) then
              shot.damage_func(shot, ava)
            end
          end
        end
      end
      --profile_section_end("shot_cycle")
    end
    for name, move in pairs(avs_that_moved_from_room) do
      avatar_moved_from_room(move.ava, move.room, move.dir)
    end
    --profile_section_start("planet_update")
    for name, planet in pairs(planets) do
      pl_update(planet)
    end
    --profile_section_end("planet_update")
  end
  gt_ticks= gt_ticks + 1
  gt_rticks= gt_rticks + 1
  if(gt_should_exit) then
    print("gt_should_exit, returning nil")
    return nil
  end
  local col_ret= collectgarbage("step", 11)
  profile_section_end("gt_idle")
  local gettime_spike= false
  if elapsed_time > (profile_section_data.gt_idle.time_total / profile_section_data.gt_idle.ticks) * 4 then
    gettime_detected_load_spikes= gettime_detected_load_spikes + 1
    gettime_spike= true
  end
  local tick_end_time= os.clock()
  elapsed_time= tick_end_time - tick_start_time
  tick_time_total= tick_time_total + elapsed_time
  ticks_passed= ticks_passed + 1
  if elapsed_time > (tick_time_total / ticks_passed) * 4 then
    os_clock_detected_load_spikes= os_clock_detected_load_spikes + 1
    if gettime_spike then
      matched_spikeds= matched_spikeds + 1
    end
  end
  return true
end

function check_wall_in_rect(rid, rect)
  local wall_in_rect= false
  local room= active_rooms[rid]
  local room_width= room.size.x
  function check_helper(x, y, other)
    if wall_in_rect then
      return
    end
    local wid= pos_to_wid(room_width, x, y)
    if room.walls[wid] then
      wall_in_rect= true
    end
  end
  local rect_pos= rect.pos
  local rect_pos_x= rect_pos.x
  local rect_pos_y= rect_pos.y
  local rect_size= rect.size
  local rect_size_x= rect_size.x
  local rect_size_y= rect_size.y
  spaced_area_iterate(rect_pos_x - rect_size_x, rect_pos_x + rect_size_x,
                      rect_pos_y - rect_size_y, rect_pos_y + rect_size_y,
                      32, check_helper)
  return wall_in_rect
end

function gt_event_handle(params, evt)
  if(evt.type == "keydown") then
    gt_keypress(evt.key)
  elseif(evt.type == "keyup") then
    gt_keyrelease(evt.key)
  end
  return nil
end

function gt_keypress(key)
  if(gt_key_down_funcs[key]) then
    gt_key_down_funcs[key]()
  else
    print("You pressed \"" .. key .. "\", for which there is not a function mapped.")
  end
end

function gt_keyrelease(key)
  if(gt_key_up_funcs[key]) then
    gt_key_up_funcs[key]()
  end
end

function gt_left_down()
  if(fly) then
    player.speed.x = player.speed.x - 8
  else
    actor.start_left_move(player)
  end
end

function gt_left_up()
  if(fly) then
    player.speed.x= 0
  else
    actor.end_left_move(player)
  end
end

function gt_right_down()
  if(fly) then
    player.speed.x = player.speed.x + 8
  else
    actor.start_right_move(player)
  end
end

function gt_right_up()
  if(fly) then
    player.speed.x= 0
  else
    actor.end_right_move(player)
  end
end

function gt_up_down()
  actor.aim_up(player)
end

function gt_up_up()
  actor.aim_mid_range(player)
end

function gt_r_down()
  rev_grav= not rev_grav
end

function gt_down_down()
  if fly then
    player.speed.y = player.speed.y + 8
  end
  actor.aim_down(player)
end

function gt_down_up()
  if fly then
    player.speed.y = 0
  end
  actor.aim_mid_range(player)
end

function gt_escape_down()
  next_play_mode= "quit"
  gt_cleanup()
end

function gt_p_down()
  collision_pause= nil
end

function gt_z_down()
  if(fly) then
    player.speed.y = player.speed.y - 8
  else
    actor.jump(player)
  end
end

function gt_z_up()
  if(fly) then
    player.speed.y = 0
  else
    player.in_jump= player.jump_max
  end
end

function gt_m_down()
  next_play_mode= "map_screen"
  gt_cleanup()
end

function gt_f_down()
  fly= not fly
end

function gt_c_down()
  collision_acoords= {}
  collision_bcoords= {}
  collision_aindices= {}
  collision_bindices= {}
  collision_ccoords= {}
  collision_cindices= {}
end

function gt_return_down()
  actor.start_firing(player)
end

function gt_return_up()
  actor.stop_firing(player)
end

function gt_space_down()
  for name, place in pairs(active_rooms[player.cur_room].places) do
    if(rects_overlap_aabb(player, place)) then
      -- End any movement the player might be in the middle of so that it isn't stuck that way when they come out.
      actor.end_left_move(player)
      actor.end_right_move(player)
      actor.stop_firing(player)
      player.speed= sc.Construct("Vec2")
      player.cur_planet= place.name
      next_play_mode= "planet"
      gt_cleanup()
    end
  end
end

function gt_screenshot_down()
  should_screenshot= true
end
