--[[
  This source file is part of TINT.
  Copyright (C) 2010 Kyzentun.

  TINT is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as
  published by the Free Software Foundation, either version 2 of the
  License, or (at your option) any later version.

  TINT is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of both the GNU General Public
  License and the GNU General Public License along with
  TINT.  If not, see <http://www.gnu.org/licenses/>.
]]

require "scui"
require "uranium"

local zoom_level= "half"
  -- "full" is zoomed in fully, each room displayed as in the minimap. (256x256 per room)
  -- "half" is each room displayed at half minimap size. (128x128 per room)
  -- "sub" is each room displayed with the subroom data. (8x8 per subroom, 64x64 per room)
  -- "sub_half" is "sub" at half scale. (4x4 per subroom, 32x32 per room)
  -- "fold" is folded view.  Nothing inside rooms is drawn.  Instead, the corners of each room
    -- are bent to try to show them folded into the shape of the hyper-polyhedron they form. (32x32 per room)
  -- "line" is similar to fold, but only lines between the centers of rooms are drawn.  (16x16 per room)
local next_zoom_in= {full= "full", half= "full", sub= "half", sub_half= "sub", fold= "sub_half", line= "fold"}
local next_zoom_out= {full= "half", half= "sub", sub= "sub_half", sub_half= "fold", fold= "line", line= "line"}
local show_dups= true
local show_only_zero_rot= false

local connection_width= 4 -- The amount of space between rooms on the map.  Used to show how the rooms are connected.
local half_conn_width= connection_width/2
local room_widths= {full= 256+connection_width, half= 128+connection_width, sub= 64+connection_width,
  sub_half= 32+connection_width, fold= 32, line= 16}

local map_center= sc.Construct("Vec2", 400, 300)

local map_widget= {}

local connected_color= {r= 0, g= .5, b= 0}
local disconnected_color= {r= .5, g= 0, b= 0}
local disco_x_edge_color= {r= 1, g= 1, b= 0}
local disco_y_edge_color= {r= 0, g= 0, b= 1}
local player_room_color= {r= 1, g= 0, b= 1}

local center_room= 1
local center_rot= 0

local scroll_steps= 8
local scroll_offsets= {up= sc.Construct("Vec2", 0, -1), right= sc.Construct("Vec2", 1, 0),
  down= sc.Construct("Vec2", 0, 1), left= sc.Construct("Vec2", -1, 0)}

local blink_cycle_mod= 0
local blink_cycle_inc= .125

-- cancer_iterate:  Iterate by growing outwards into unfilled space.
-- positions is an array of the valid positions.
-- start_position is the id of the position to start at.
-- start_info is the parent information to pass to iterate_function the first time it is called.
-- iterate_function is a function to call for each position that is grown into.
-- iterate_function should take its position and information from its parent as args.
-- iterate_function must return a table of new positions to grow into.  Each element of this table must be a table
--   containing the id of the new position, the new position, and the information to pass to iterate_function when
--   it is called for the new position.
function cancer_iterate(positions, start_position, start_info, iterate_function)
  local active_positions= {{pos= positions[start_position], info = start_info}}
  local filled_position= {[start_position]= true}
  while #active_positions >= 1 do
    local growth_positions= {}
    for ap= 1, #active_positions, 1 do
      local new_growth= iterate_function(active_positions[ap].pos, active_positions[ap].info)
      if new_growth then
        for ng= 1, #new_growth, 1 do
          local cng= new_growth[ng]
          local cng_pos= cng.pos
          if not filled_position[cng.id] and positions[cng.id] then
            if positions[cng.id].x == cng_pos.x and positions[cng.id].y == cng_pos.y then
              filled_position[cng.id]= true
              growth_positions[#growth_positions+1]= cng
            end
          end
        end
      end
    end
    active_positions= growth_positions
  end
end

function calc_needed_dimensions(rw, sw, sh)
  local rooms_across_width= math.ceil(sw / rw)
  local rooms_across_height= math.ceil(sh / rw)
  if rooms_across_width % 2 == 0 then
    rooms_across_width= rooms_across_width + 1
  end
  if rooms_across_height % 2 == 0 then
    rooms_across_height= rooms_across_height + 1
  end
  local needed_width= rooms_across_width * rw
  local needed_height= rooms_across_height * rw
  return needed_width, needed_height, rooms_across_width, rooms_across_height
end

function detail_render(window, dest, dw, dh, center, room_width, scale, file_suffix)
  dest:SetPrimitiveColor(0, 0, 0)
  dest:DrawRect(0, 0, dw, dh)

  local positions= {}
  local spacing= room_width
  local rid_map= {}
  print("Room positions:")
  local rooms_across_width= math.ceil(dw / spacing)
  local left_bound= 0
  local right_bound= dw
  local top_bound= 0
  local bottom_bound= dh

  function rend_pos_to_id(x, y)
    return pos_to_id(left_bound, top_bound, rooms_across_width, spacing, x, y)
  end

  function pos_adder(x, y, other)
    local cur_id= rend_pos_to_id(x, y)
    positions[cur_id]= sc.Construct("Vec2", x, y)
    print(cur_id..":  "..x..", "..y)
  end

  -- Subtract 1 from the right and bottom bounds and use no_edge_spaced_area_iterate so that we don't get the false
  -- positions at the edge of the area.
  no_edge_spaced_area_iterate(left_bound, right_bound-1, top_bound, bottom_bound-1, spacing, pos_adder)
  local start_id= rend_pos_to_id(center.x, center.y)
  print("Starting at position "..start_id)
  local start_info= {rid= center_room, rot= center_rot, pos_id= start_id}
  local rendered_rooms= {}
  function room_render(pos, info)
    local rmx= math.floor((pos.x - left_bound) / spacing)
    local rmy= math.floor((pos.y - top_bound) / spacing)
    print("Rendering room "..info.rid.." at "..pos.x..", "..pos.y.."  ("..rmx..", "..rmy..")  pos;  "..rend_pos_to_id(pos.x, pos.y))
    local rot= info.rot
    if not rid_map[rmx] then
      rid_map[rmx]= {}
    end
    rid_map[rmx][rmy]= {rid= info.rid, rot= info.rot, pos= pos}
    local room= active_rooms[info.rid]
    local had_to_load= false
    if not room then
      load_room(info.rid)
      had_to_load= true
    end
    room= active_rooms[info.rid]
    if room then
      local room_mini= SCUtil.ScaleFast(load_img_managed("rooms/"..info.rid..file_suffix..".png"), 32*scale, 32*scale)

      local growth_info= {}
      local offsets= {}
      if rot == 0 then
        dest:Blit(room_mini, pos.x+(half_conn_width), pos.y+(half_conn_width))
        rid_map[rmx][rmy].neighbors= {left= room.conns.left.side, right= room.conns.right.side,
                                      top= room.conns.top.side, bottom= room.conns.bottom.side}
        offsets= { left= sc.Construct("Vec2", -room_width, 0),
                   right= sc.Construct("Vec2", room_width, 0),
                   top= sc.Construct("Vec2", 0, -room_width),
                   bottom= sc.Construct("Vec2", 0, room_width) }
      elseif rot == 1 then
        dest:Blit(SCUtil.RotateRight(room_mini), pos.x+(half_conn_width), pos.y+(half_conn_width))
        rid_map[rmx][rmy].neighbors= {left= room.conns.bottom.side, right= room.conns.top.side,
                                      top= room.conns.left.side, bottom= room.conns.right.side}
        offsets= { left= sc.Construct("Vec2", 0, -room_width),
                   right= sc.Construct("Vec2", 0, room_width),
                   top= sc.Construct("Vec2", room_width, 0),
                   bottom= sc.Construct("Vec2", -room_width, 0) }
      elseif rot == 2 then
        dest:Blit(SCUtil.Flip(room_mini), pos.x+(half_conn_width), pos.y+(half_conn_width))
        rid_map[rmx][rmy].neighbors= {left= room.conns.right.side, right= room.conns.left.side,
                                      top= room.conns.bottom.side, bottom= room.conns.top.side}
        offsets= { left= sc.Construct("Vec2", room_width, 0),
                   right= sc.Construct("Vec2", -room_width, 0),
                   top= sc.Construct("Vec2", 0, room_width),
                   bottom= sc.Construct("Vec2", 0, -room_width) }
      elseif rot == 3 then
        dest:Blit(SCUtil.RotateLeft(room_mini), pos.x+(half_conn_width), pos.y+(half_conn_width))
        rid_map[rmx][rmy].neighbors= {left= room.conns.top.side, right= room.conns.bottom.side,
                                      top= room.conns.right.side, bottom= room.conns.left.side}
        offsets= { left= sc.Construct("Vec2", 0, room_width),
                   right= sc.Construct("Vec2", 0, -room_width),
                   top= sc.Construct("Vec2", -room_width, 0),
                   bottom= sc.Construct("Vec2", room_width, 0) }
      else
        print("Recieved out of range rotation "..rot)
      end
      for k, v in pairs(room.conns) do
        if room.conns[k] then
          local pos_plus= pos + offsets[k]
          if pos_plus.x >= left_bound and pos_plus.x <= right_bound and
             pos_plus.y >= top_bound and pos_plus.y <= bottom_bound then
            if show_dups or not rendered_rooms[room.conns[k].side] then
              if not show_only_zero_rot or room.conns[k].rot == 0 then
                rendered_rooms[room.conns[k].side]= true
                growth_info[#growth_info+1]= {
                  id= rend_pos_to_id(pos_plus.x, pos_plus.y),
                  pos= pos_plus,
                  info= {rid= room.conns[k].side, rot= (room.conns[k].rot + rot) % 4} }
              end
            end
          end
        end
      end
      return growth_info
    else
      print("Room is not active, unable to render.")
    end
    return nil
  end
  cancer_iterate(positions, start_id, start_info, room_render)
  for x, col in pairs(rid_map) do
    for y, el in pairs(col) do
      if el.rid == player.cur_room then
        set_cholera(dest, player_room_color)
        dest:DrawBox(el.pos.x, el.pos.y, room_width, room_width, half_conn_width)
      else
        set_cholera(dest, disconnected_color)
        if rid_map[x-1] then
          if rid_map[x-1][y] then
            if rid_map[x-1][y].rid == el.neighbors.left then
              set_cholera(dest, connected_color)
            end
          else
            set_cholera(dest, disco_y_edge_color)
          end
        else
          set_cholera(dest, disco_x_edge_color)
        end
        dest:DrawRect(el.pos.x, el.pos.y+half_conn_width, half_conn_width, room_width-connection_width)
        set_cholera(dest, disconnected_color)
        if rid_map[x+1] then
          if rid_map[x+1][y] then
            if rid_map[x+1][y].rid == el.neighbors.right then
              set_cholera(dest, connected_color)
            end
          else
            set_cholera(dest, disco_y_edge_color)
          end
        else
          set_cholera(dest, disco_x_edge_color)
        end
        dest:DrawRect(el.pos.x+(room_width-half_conn_width), el.pos.y+half_conn_width, half_conn_width, room_width-connection_width)
        set_cholera(dest, disconnected_color)
        if rid_map[x] then
          if rid_map[x][y-1] then
            if rid_map[x][y-1].rid == el.neighbors.top then
              set_cholera(dest, connected_color)
            end
          else
            set_cholera(dest, disco_y_edge_color)
          end
        else
          set_cholera(dest, disco_x_edge_color)
        end
        dest:DrawRect(el.pos.x+half_conn_width, el.pos.y, room_width-connection_width, half_conn_width)
        set_cholera(dest, disconnected_color)
        if rid_map[x] then
          if rid_map[x][y+1] then
            if rid_map[x][y+1].rid == el.neighbors.bottom then
              set_cholera(dest, connected_color)
            end
          else
            set_cholera(dest, disco_y_edge_color)
          end
        else
          set_cholera(dest, disco_x_edge_color)
        end
        dest:DrawRect(el.pos.x+half_conn_width, el.pos.y+(room_width-half_conn_width), room_width-connection_width, half_conn_width)
      end
    end
  end
end

function full_render(dest)
  local w, h= dest:GetSize()
  detail_render(window, dest, w, h, map_center, room_widths.full, 8, "")
end

function half_render(dest)
  local w, h= dest:GetSize()
  detail_render(window, dest, w, h, map_center, room_widths.half, 4, "")
end

function sub_render(dest)
  local w, h= dest:GetSize()
  detail_render(window, dest, w, h, map_center, room_widths.sub, 2, "sub")
end

function sub_half_render(dest)
  local w, h= dest:GetSize()
  detail_render(window, dest, w, h, map_center, room_widths.sub_half, 1, "sub")
end

function fold_render(dest)
  dest:SetPrimitiveColor(.5, .5, 1)
  dest:DrawRect(self._x+1, self._y+1, self.w-2, self.h-2)
end

function line_render(dest)
  dest:SetPrimitiveColor(0, 0, 1)
  dest:DrawRect(self._x+1, self._y+1, self.w-2, self.h-2)
end

local render_funcs= {full= full_render, half= half_render, sub= sub_render,
  sub_half= sub_half_render, fold= fold_render, line= line_render}

function map_widget:outline_render()
  if(self.outline_color) then
    set_cholera(self.cur_img, self.outline_color)
  else
    screen:SetPrimitiveColor(1, 1, 1)
  end
  screen:DrawBox(self._x, self._y, self.w, self.h)
  if(self.inside_color) then
    set_cholera(screen, self.inside_color)
  else
    screen:SetPrimitiveColor(0, 0, 0)
  end
  screen:DrawRect(self._x+1, self._y+1, self.w-2, self.h-2)
  screen:SetClipRect(self._x+1, self._y+1, self.w-2, self.h-2)
end

function map_widget:RenderSelf(window)
  print("Rendering in "..zoom_level.." mode.")
  self:outline_render()
  screen:Blit(self.cur_img, -self.cur_offset.x, -self.cur_offset.y, self.w, self.h, 0, 0, self.cur_alpha)
  if self.scroll_to_img then
    screen:Blit(self.scroll_to_img, -self.scroll_to_offset.x, -self.scroll_to_offset.y, self.w, self.h, 0, 0, self.scroll_to_alpha)
  end
  screen:Update(self._x, self._y, self.w, self.h)
end

function map_widget:fade_to_new_state()
  local room_width= room_widths[zoom_level]
  local needed_width, needed_height= calc_needed_dimensions(room_width, self.w, self.h)
  local x_margin= math.ceil((self.w - needed_width) / 2)
  local y_margin= math.ceil((self.h - needed_height) / 2)
  self.scroll_to_img= sc.Construct("Graphic", needed_width, needed_height)
  render_funcs[zoom_level](self.scroll_to_img)
  self.scroll_to_offset= sc.Construct("Vec2", x_margin, y_margin)
  self.offset_step= sc.Construct("Vec2")
  self.scroll_to_alpha= 0
  self.alpha_step= 1 / scroll_steps
  local scx= self.x + (self.w / 2)
  local scy= self.y + (self.h / 2)
  self.cursor_x= scx - (room_width/2)
  self.cursor_y= scy - (room_width/2)
  self.cursor_w= room_width
  self.cursor_h= room_width
  self.cursor_t= half_conn_width
end

function map_widget_reinit(ret)
  local room_width= room_widths[zoom_level]
  local needed_width, needed_height= calc_needed_dimensions(room_width, ret.w, ret.h)
  local x_margin= math.ceil((ret.w - needed_width) / 2)
  local y_margin= math.ceil((ret.h - needed_height) / 2)
  ret.cur_img= sc.Construct("Graphic", needed_width, needed_height)
  render_funcs[zoom_level](ret.cur_img)
  ret.cur_offset= sc.Construct("Vec2", x_margin, y_margin)
  ret.cur_alpha= 1
  local scx= ret.x + (ret.w / 2)
  local scy= ret.y + (ret.h / 2)
  ret.cursor_x= scx - (room_width/2)
  ret.cursor_y= scy - (room_width/2)
  ret.cursor_w= room_width
  ret.cursor_h= room_width
  ret.cursor_t= half_conn_width
  return ret
end

function map_widget_init()
  local ret= {x= 0, y= 0, w= screen_height, h= screen_height, type= map_widget,
    hotkeys= {["-"]=true, ["="]= true, d= true, r= true, up= true, down= true, left= true, right= true},
    inkey= "=", outkey= "-", dupkey= "d", rotkey= "r", upkey= "up", downkey= "down", leftkey= "left", rightkey= "right"}
  map_center= sc.Construct("Vec2", ret.x+(ret.w/2), ret.y+(ret.h/2))
  return map_widget_reinit(ret)
end

function map_widget:idle()
  if self.scroll_to_img then
    self.cur_offset= self.cur_offset + self.offset_step
    self.scroll_to_offset= self.scroll_to_offset + self.offset_step
    self.cur_alpha= self.cur_alpha - self.alpha_step
    self.scroll_to_alpha= self.scroll_to_alpha + self.alpha_step
    if self.scroll_to_alpha >= 1 then
      self.cur_img= self.scroll_to_img
      local room_width= room_widths[zoom_level]
      local needed_width, needed_height= calc_needed_dimensions(room_width, self.w, self.h)
      local x_margin= math.ceil((self.w - needed_width) / 2)
      local y_margin= math.ceil((self.h - needed_height) / 2)
      self.cur_offset= sc.Construct("Vec2", x_margin, y_margin)
      self.cur_alpha= 1
      self.scroll_to_img= nil
      self.scroll_to_offset= nil
      self.scroll_to_alpha= nil
    end
    scui.RenderWidget(self)
  end
  screen:SetPrimitiveColor(1 * blink_cycle_mod, .25 * blink_cycle_mod, 0)
  screen:DrawBox(self.cursor_x, self.cursor_y, self.cursor_w, self.cursor_h, self.cursor_t)
  blink_cycle_mod= blink_cycle_mod + blink_cycle_inc
  if blink_cycle_mod > 1 then
    blink_cycle_inc= -.125
  end
  if blink_cycle_mod < .25 then
    blink_cycle_mod= .25
    blink_cycle_inc= .125
  end
  screen:Update(self.cursor_x, self.cursor_y, self.cursor_w, self.cursor_h)
end

function map_widget:OnHotKey(key, mod)
  print("Picked up hotkey.")
  if(not key) then print ("map_widget:OnHotKey key is nil") end
  if(not mod) then print ("map_widget:OnHotKey mod is nil") end
  print("list_widget:OnHotKey:  key = " .. key)
  if key == self.inkey then
    zoom_in()
  elseif key == self.outkey then
    zoom_out()
  elseif key == self.quitkey then
    map_screen_cleanup()
  elseif key == self.dupkey then
    no_dup_toggle()
  elseif key == self.rotkey then
    show_only_zero_rot_toggle()
  elseif key == self.upkey then
    self:shift_center("up")
  elseif key == self.rightkey then
    self:shift_center("right")
  elseif key == self.downkey then
    self:shift_center("down")
  elseif key == self.leftkey then
    self:shift_center("left")
  end
end

function map_widget:OnMouseDown()
end

function zoom_in()
  print("Zooming in.")
  zoom_level= next_zoom_in[zoom_level]
  map_screen_widget[1]:fade_to_new_state()
end

function zoom_out()
  print("Zooming out.")
  zoom_level= next_zoom_out[zoom_level]
  map_screen_widget[1]:fade_to_new_state()
end

function no_dup_toggle()
  print("Toggling displaying duplicate rooms.")
  show_dups= not show_dups
  map_screen_widget[1]:fade_to_new_state()
end

function show_only_zero_rot_toggle()
  show_only_zero_rot= not show_only_zero_rot
  map_screen_widget[1]:fade_to_new_state()
end

function map_widget:shift_center(dir_name)
  local dir_index= (dir_name_to_index[dir_name] - center_rot) % 4
  local side_name= dir_index_to_name[dir_index]
  print("Attempting to shift map \""..dir_name.."\".  dir_index:  "..dir_index.."  side_name:  "..side_name)
  local conn= active_rooms[center_room].conns[side_name]
  if conn then
    center_room= conn.side
    center_rot= (center_rot + conn.rot) % 4
    
    local room_width= room_widths[zoom_level]
    local needed_width, needed_height= calc_needed_dimensions(room_width, self.w, self.h)
    self.scroll_to_img= sc.Construct("Graphic", needed_width, needed_height)
    render_funcs[zoom_level](self.scroll_to_img)
    self.scroll_to_offset= self.cur_offset + (scroll_offsets[dir_name] * room_width)
    self.offset_step= scroll_offsets[dir_name] * (room_width / scroll_steps * -1)
    self.scroll_to_alpha= 0
    self.alpha_step= 1 / scroll_steps
  else
    print("No connection.")
  end
end

function map_screen_cleanup()
  return "Leaving."
end

function map_screen_idle()
  map_screen_widget[1]:idle(map_screen_widget[1])
end

function map_screen_start()
  print("Building map_screen_widget.")
  center_room= player.cur_room
  center_rot= player.cur_rot
  map_screen_widget= {
  x= 0, y= 0, w= screen_width, h= screen_height,
  type= uranium.box,
  filled= true,
  map_widget_init(),
  -- Kind of a hack, need to figure out why the exit through map_widget:OnHotKey doesn't work.
  {x= 800, y= 600, w= 0, h= 0, type= uranium.button, text= " ", action= map_screen_cleanup, hotkeys= {["escape"]= true}}
  }
  print("Compiling widget.")
  scui.CompileWidget(map_screen_widget)
  print("Running modal.")
  scui.RunModal(map_screen_widget, {idle= map_screen_idle}, uranium.cursor())
  -- RunModal returns when one of the functions attached to part of the dialog returns non-nil.
  print("Map screen finished.")
  map_screen_widget= nil
  next_play_mode= "travel"
end
